/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */
 
//  xxxxxxxxxxxxx          version: 1.0              xxxxxxxxxxxxxxx
		 
//  xxxxxxxxxxxxxxx   global variable declarations   xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var item_cnt= 5;       // The number of stimuli in the task

var rep_cnt = 3;       // maximum number of sounds repetitions (first + second tone) in each item in task

var before = 10;       // Delay time before play first tone-sound x (clock = 100ms) = ms

var after = 10;				 // Delay time before play second tone-sound x clock = ms	
  
	// var range = '0'  => see stimGen()

var is_back = true;    // if (is_back==true) => user can go back to previous task/item.  

var is_req = false;    // if (is_req==true) => user must selected one option before moving to a new item.

var prg_type = true;   //  type of information in Progress Bar (true => Items, false => Nothing) 
    		
var demo = false;     // if (demo == true) => Show task solutions 

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

 
// Other global constants and variables:

const clock = 100;  // basic timer resolution in ms. Default = 100ms

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var step = -1;   		 	 // global timer counter => see appTimer()

var first = false;    // true => play first tone (sound); false => play second tone (sound)
   
var dirty = false;  // if  (dirty==true) => user changed any options in form 

var startDate;  // Task start time

var item = 0;   // current item in task     

var delay;     // timer will wait for this period. delay == before (if first == true) or after (if first == false)

// global arrays
	
var stim = [];   // stimulus content (length/name of .wav files in audio dir) for each item. range dimensions: [0..item_cnt-1] x [0..1]

var ans = [];   //  the user's response to each item in the task [0..item_cnt-1]

var cor = [];    // solutions for each item in the task  [0..item_cnt-1]

var repeated = [];  // for each item: the number of completed repetitions of the sound stimulus for each item. [0..item_cnt-1]



// application init [create/start Timer, check storage, show start panel..] 
function init() 
{  
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);  
    
  is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }		
		
	setIns();		
	showStart(); 
}


// application timer handler function
// basic timer resolution is aprox 100 ms
function appTimer()    
{    
	 if (step >= 0)  //   if (step < 0 => timer nop, idle, just wait
	 {          		
		 step++;  
		 if (step == delay) 
		 {  
	     step = -1;       // timer must wait. audio play in progress.  nop, idle, do none, Timer no operation  			   
			 playSound();     // after delay draw second part of stimulus (stimulus after delay)            
		 }
	 }
} 	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 and Edge offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		
   
  if (rep_cnt > 1) {		
		document.getElementById('ins_rep_cnt').style.display = 'block';
		document.getElementById('ins_rep_cnt').textContent= 'If you did not hear the tones in the current task, click the "Repeat" button and both tones will play again. You can use this option ' + rep_cnt + ' times within the same task.';
	}
	else  {
		document.getElementById('ins_rep_cnt').style.display = 'none';
	}
  document.getElementById('ins_req').style.display = is_req ? 'block' : 'none';  
	document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';	
	
}	

 
// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'item_cnt':
		case 'rep_cnt':
		case 'before':
		case 'after':
			if (e.validity.valid)  {
				msg.textContent= ''; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;  		
		
    case 'range': 
      switch (e.selectedIndex)
			{
				case 0:
				  msg.textContent= 'All available tones will be used. The probability of repeating the same stimuli within the task is the lowest.';
					break;
				case 1:
				  msg.textContent= 'The difference in the length of the first and second tones within each item will be within the range 0..100 ms (lower and upper bound are included)';
					break;
				case 2:
				  msg.textContent= 'The difference in the length of the first and second tones within each item will be within the range 100..500 ms (lower and upper bound are included)';
					break;
				case 3:
				  msg.textContent= 'The difference in the length of the first and second tones within each item will be within the range 500..1000 ms (lower and upper bound are included)';
					break;
				case 4:
				  msg.textContent= 'The difference in length of the first and second tones within each item will be greater than or equal to 1 second';
					break;					
			}     			
      break; 		
		
		
    case 'is_back':     		
			msg.textContent= e.checked ? 'Yes' : 'No. Returning to previous items is not allowed.';    
			break;
    
    case 'is_req':
    	msg.textContent= e.checked ? 'You must select one of the options before moving on to the next item.' : 'It is not necessary to select one of the options before moving on to the next item.';    
			break;
    
    case 'prg_type':                                                                                                                                                  
      msg.textContent= e.checked ? 'Progress bar will show the number of the current item and the number of remaining items until the end of the task.' : 'Progress bar will be invisible';     
      break;    
							
		case 'demo':
       msg.textContent= e.checked ? 'The correct solution will be displayed on each item.' : 'No';  
       break; 
		
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty == true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity.valid object in all input type="number"
function optionsValid()  
{     
  let res= document.getElementById('item_cnt').validity.valid && document.getElementById('rep_cnt').validity.valid
				&& document.getElementById('before').validity.valid && document.getElementById('after').validity.valid;	
  return res;              
}


// start task (research) 
function startTask(e)   
{     
  initLayout();  
	
	setPar();
 
	stimGen();
		   
	showItem();     // show div/panel for first item
   		
}


// init page layout when task is started. init visible/invisible html dom elements, progress bar
function initLayout()
{
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
  document.getElementById('task_section').style.display= 'block'; 
	
	document.getElementById('nav_items').textContent= '';	
  if (!prg_type)  {
		document.getElementById('progress_holder').style.display= 'none';    	
	}	
	else  {
		document.getElementById('progress_holder').style.display= 'block'; 
		document.getElementById('progress').setAttribute('style', 'width:100%');    				   			
		document.getElementById('progress').setAttribute('style', 'width:100%');  // reset progress bar width						 	
	} 
  document.getElementById('solution').style.display= demo ? 'block' : 'none';	
}


// set internal application research parameters (global vars)
function setPar()  
{  
  stim= []; 
  ans= [];
  cor = [];   	
	repeated= [];
	for (let i= 0; i< item_cnt; i++)   {  // i => item        
    ans[i]= -1;	 // user is not selected any radio choice (1, 2, =)
		repeated[i]= 0;
  } 
  item= 0;    	
  startDate = new Date();			
	first= true; 
	delay= before; 
}


// generates task stimuli. stimulus generator. set correct choices
function stimGen()	
{
	const files= [ 0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2, 2.5, 3, 3.5, 4, 4.5, 5 ];
  let len= files.length, f, s, d, ok, min_d, max_d;   
	let is_range= (document.getElementById('range').selectedIndex > 0);
	if (is_range) 
	{
		switch (document.getElementById('range').selectedIndex)
		{
			case 1: // 0..100 ms
			  min_d= 0;
				max_d= 0.1;
				break;
			case 2:  // 100..500 ms
			  min_d= 0.1;
				max_d= 0.5;
				break;
			case 3:  // 500..1000 ms
			  min_d= 0.5;
				max_d= 1;
				break;
			case 4:  // >= 1000 ms
			  min_d= 1;
				max_d= 10;
				break;				
		}
	}
	for (let i= 0; i< item_cnt; i++)   
	{  // i => item  
		stim[i]= [];
		if (is_range)		
		{	
			do  
			{				
				f= files[Math.floor(Math.random() * len)];   // index returns a random integer from 0 to files.length-1  (first audio file)
				s= files[Math.floor(Math.random() * len)];   // second audio file 				
				if (i == 0) { 
					ok= true; 
				}
				else	{  // 1. condition: 2 files in i-iteration must be different from 2 files in i-1 iteration
					ok= ( f!= stim[i-1][0] && f != stim[i-1][1] && s != stim[i-1][0] && s != stim[i-1][1] );  
				}			  
				d= Math.abs(f - s).toFixed(2);  // must be toFixed()
				ok = ok && ( f == s || (d >= min_d && d <= max_d) );  // condition |f-a| must in range [min_d .. max_d] or two tones are equal				
			} while (!ok);
		}
		else  
		{
			do 
			{
				f= files[Math.floor(Math.random() * len)];   // index returns a random integer from 0 to files.length-1  (first audio file)
				s= files[Math.floor(Math.random() * len)];   // second audio file 
				if (i == 0) { 
					ok= true; 
				}
				else	{  // only this condition: 2 files in i-iteration must be different from 2 files in i-1 iteration
					ok= ( f!= stim[i-1][0] && f != stim[i-1][1] && s != stim[i-1][0] && s != stim[i-1][1] );  
				}	        
			} while (!ok); 
		}
		stim[i][0]= f;  // first tone (audio wav file)
		stim[i][1]= s;  // second tone (audio wav file)
		// set correct solution for each item in task..
		if ( stim[i][0] > stim[i][1] ) {
		  cor[i]= 1;	// The first tone lasted longer
		}
		else if ( stim[i][0] < stim[i][1] ) {
		  cor[i]= 2;	// The second tone lasted longer
		}
		else  {
		  cor[i]= 0;	// The two tones (sounds) were of equal length
		}
   }	  
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
	
	if (document.getElementById('item_cnt').validity.valid)  {  
	  item_cnt= parseInt(document.getElementById('item_cnt').value, 10);			
  } 
	 
  if (document.getElementById('rep_cnt').validity.valid) {
		rep_cnt= parseInt(document.getElementById('rep_cnt').value, 10);	
  }
	 
  if (document.getElementById('before').validity.valid) {
		before= Math.round(parseInt(document.getElementById('before').value, 10)/clock);
  }
  
	if (document.getElementById('after').validity.valid) {
		after= Math.round(parseInt(document.getElementById('after').value, 10)/clock);  
  }

	is_back= document.getElementById('is_back').checked ? true: false;       
  
  is_req= document.getElementById('is_req').checked ? true: false; 
	
	prg_type= document.getElementById('prg_type').checked ? true : false;  // display item count on footer or nothing
	
  demo= document.getElementById('demo').checked ? true: false;	
	
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
  
}


// goto previous item (stimulus)
function priorItem()  
{  
  if (item > 0)  {
    item--;  
    showItem();     // draw again previous item 
  }         
}


// goto next item (stimulus)
function nextItem()  
{ 
  if (is_req && ans[item] == - 1) {
		alert('You must select one of the options (1, 2, =) before moving on to the next item.');
		return;
	} 
	item++;					
	if (item < item_cnt)  {         		 
    showItem();     // show next item => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


// show current item_div panel/div. items are in range: 0..item_cnt-1     
function showItem() 
{    		
  // update footer (navigation buttons, progress bar,..)
	document.getElementById('prior').style.display= (item > 0 && is_back) ? 'inline': 'none';  // set footer navigation prior button (left arrow) 		
	 
	let btn= document.getElementById('next');     
	if(item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update "Next item" image on "right arrow" button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Next';
	} 
					
	if (prg_type)  {    // update Progress bar => show current item count and remaining item count on Progress Bar (default)                                                                                                                                                       
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent= 'Item= ' + (item + 1) + ' Remaining= ' + (item_cnt - item-1);  
	}
	
	document.getElementById('repeat').style.display= repeated[item] < rep_cnt ? 'block' : 'none';
	
	// update user selected choice (1, 2, =)
	for (let i= 0; i<=2; i++) {
		document.getElementById(i.toString()).checked= ans[item] == i ? true: false;
	}
		
	if (demo)  
	{		
		if ( cor[item] == 1 ) {
		  document.getElementById('solution').textContent= 'The first tone lasted longer';
		}
		else if ( cor[item] == 2 ) {
		  document.getElementById('solution').textContent= 'The second tone lasted longer';
		}
		else  {
		  document.getElementById('solution').textContent= 'The two tones were of equal length';
		}		 
	}
		
	if (repeated[item] == 0)  {  // only first time when item is displayed	
	  document.getElementById('item_div').style.display= 'none';
		step= 0;   // "enable" timer. timer is on
	}
  
}


// play only one sound stimulus. play audio from ../audio/fix dir. 
function playSound()
{   
	let sound= document.getElementById('sound');		
  let file = first ? 'audio/' + stim[item][0] + '.wav' : 'audio/' + stim[item][1] + '.wav';
	sound.setAttribute('src', file);  // sound.load();
	sound.play();
	sound.onended = function() {
		audioEnded();    
	}; 			  
} 


// audio ended => when the audio file is finished
function audioEnded()
{  
	first = !first;
	delay= first ? before : after;   // update timer delay period
  if ( first )  {  // second tone ended 
		step= -1;  // "disable" timer nop, idle
		repeated[item]++;
		document.getElementById('item_div').style.display= 'block';		
		document.getElementById('repeat').style.display= repeated[item] < rep_cnt ? 'block' : 'none';
	}
	else  {    		
		step= 0;  // enable timer only if second tone must be played after delay period
	}
} 


// play again first and second tone in current item.
// this event is trigger by user click
function repeatSound()
{
	document.getElementById('item_div').style.display= 'none';
	first= true; 
	step= 0;   // activate timer again. timer is on
}


// evaluate user choice (3 radio html elements)
// v => value [ 1 = The first tone lasted longer; 2 = The second tone lasted longer; 0 = The two tones were of equal length ] 
function scoreItem(v)
{ 
	let e= document.getElementById(v);  // radio html object/elem 
  v= parseInt(v, 10);
	if ( ans[item] == v )  {  // toogle choice.  2 x click on same radio will reset answer 
		ans[item]= -1;
    e.checked= false;		
	}
  else  {			
		ans[item]= v;
		e.checked= true;
	}		
}


// end of task (research)
function theEnd()  
{   
	step= -1;  // timer nop
	
	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
  document.getElementById('result_wrapper').style.display= 'block';
	
	let endDate= new Date();
	let task_time= Math.round( (endDate.getTime() - startDate.getTime() ) / 1000);  // seconds
	
	// points = sum of scores for each item in task. For each item: if the item was successfully completed = 1 point. Else = 0 points.
  // err = number of errors in task. mis = number of missing items in task 
	let points = 0, err= 0, mis= 0, rows = [], header= ['Item', 'Choice', 'Correct', 'Point'];    
	for (let i= 0; i< item_cnt; i++)   
	{
    let p = 0;  // score only for current item (0 or 1)
		let v = ans[i]; // recoded user choice
		if ( ans[i] == -1 )  {
		  mis++;
			v= '';  // recode user choice
    }		
		else if ( ans[i] == cor[i]  ) {  
      p= 1;			
    } 
    else  {
      err++;
    }
    points+= p;		
    let c = cor[i] == 0  ? '=' : cor[i];  // recoded correct choice
    if ( ans[i] == 0 ) {
			v= '=';  // recode user choice 0 => =
		}    		
		rows[i]= [ parseInt(i+1,10), v, c, p ];  // => ['Item', 'Choice', 'Correct', 'Point']					
	}		
	document.getElementById('score').textContent= 'Sum of points= ' + points + '. Maximum points= ' + item_cnt + '. Errors= ' + err + '.  Missing= ' + mis + '.'; 
	
  rows.push( [ 'Sum', '', '', points ] ); 
	createTableArray(document.getElementById('table1'), rows, header);	
	
	// "chart" => repaint "chart" 
  let ok= Math.round(100 * points / item_cnt).toFixed(2);
	let not_ok= Math.round(100 - ok).toFixed(2); 
	document.getElementById('chart_1').setAttribute('style', 'width:' + ok + '%');    // !important
	document.getElementById('chart_2').setAttribute('style', 'width:' + not_ok + '%');	
	document.getElementById('chart').title= 'The percentage of points scored (green) = ' + ok + '% (100*' + points + '/' + item_cnt + ')';  	
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
		
	try 
  {
    if (store) {
			setStorage(points, mis, task_time);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }   	 
	
}

 
/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:	 
	 points = total number of points in task. each item => 0 or 1 point
   missing = 	missing values in task (no selected choice) 
	 t_time => total task time in seconds */
function setStorage(points, missing, t_time)   
{   
  let sel= document.getElementById('range'); 
	let rng= sel.selectedIndex > 0 ? sel.options[sel.selectedIndex].text : ''; 

	let oldItems = JSON.parse(localStorage.getItem('wm18-v1.0')) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
      'points' : points,
			'missing' : missing,	
			'items': item_cnt,			
			'demo' : demo ? 1 : 0,			
			'repeated': rep_cnt,			
	    'time-before' : before * clock,        
      'time-after' : after * clock,			
			'range': rng,
			'backward': is_back ? 1 : 0,
			'required' : is_req ? 1 : 0,
      'task-time': t_time			
  };

  oldItems.push(newItem);

  localStorage.setItem('wm18-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);   
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined)	{
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').setAttribute('class', 'center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';        //.removeAttribute("disabled");  
  document.getElementById('hide_option').style.display = 'none';   
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
}


// begin of new code for export localStorage app data:  data= localStorage.getItem('myApp-v1.0');

// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (document.getElementById('store_group').disabled === false && localStorage.getItem('wm18-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('wm18-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' WM18 results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';   // xxx
		}
	}
  dirty= false;
  document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('wm18-v1.0') ? JSON.parse(localStorage.getItem('wm18-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no WM18 data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' WM18 results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved WM18 results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 {  
   window.localStorage.removeItem('wm18-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('wm18-v1.0');
	if (!data)  {
    alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('wm18-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object id (div or table)
function newWinLast(id)
{ 
  let e = document.getElementById(id);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{     
  let tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
